/*
 *  This file is part of Siena, a wide-area event notification system.
 *  See http://www.cs.colorado.edu/serl/siena/
 *  
 *  Author: Antonio Carzaniga <carzanig@cs.colorado.edu>
 *  See the file AUTHORS for full details. 
 *  
 *  Copyright (C) 1998-2002 University of Colorado
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
 *  USA, or send email to serl@cs.colorado.edu.
 *  
 *  
 *  $Id: sff_parser.yy,v 1.2 2002/11/26 18:50:53 carzanig Exp $
 */

%{
#include <cstring>
#include <iostream>
#include <string>
#include <set>
#include <map>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#include <sys/resource.h>

#if HAVE_UNISTD_H
# include <sys/types.h>
# include <unistd.h>
#else
# error you need unistd.h
#endif

#include "yysfftypes.h"
#include <siena/forwarding.h>
#include <siena/allocator.h>

#include "simple_fwd_types.h"

#define YYERROR_VERBOSE
#ifdef YYDEBUG
#undef YYDEBUG
#endif

int yyerror(char *);

/** resource meter
 *
 *  measures CPU usage for both user and system time
 **/
class Timer {
public:
    void start();
    void stop();

    long elapsed_sys_msec();
    long elapsed_usr_msec();

private:
    struct rusage s_start;
    struct rusage s_stop;
};

inline void Timer::start() {
    getrusage(RUSAGE_SELF, &s_start);
}

inline void Timer::stop() {
    getrusage(RUSAGE_SELF, &s_stop);
}

inline long Timer::elapsed_usr_msec() {
    long secs = s_stop.ru_utime.tv_sec - s_start.ru_utime.tv_sec;
    long msec = (s_stop.ru_utime.tv_usec - s_start.ru_utime.tv_usec) / 1000;
    return secs * 1000 + msec;
}

inline long Timer::elapsed_sys_msec() {
    long secs = s_stop.ru_stime.tv_sec - s_start.ru_stime.tv_sec;
    long msec = (s_stop.ru_stime.tv_usec - s_start.ru_stime.tv_usec) / 1000;
    return secs * 1000 + msec;
}

class ICounter : public MatchHandler {
public:
    ICounter() : c(0) { }
    virtual bool output(if_t);

    unsigned long get_count() { return c; }
private:
    unsigned long  c;
};

bool ICounter::output(if_t x) {
    ++c;
    return false;
}

class IPrinter : public ICounter {
public:
    IPrinter(): ICounter() {};
    virtual bool output(if_t);
    void flush(ostream & os);

 private:
    set<if_t> iset;
};

bool IPrinter::output(if_t x) {
    ICounter::output(x);
    iset.insert(x);
    return false;
}

void IPrinter::flush(ostream & os) {
    os << "->";
    set<if_t>::const_iterator i;
    for(i = iset.begin(); i != iset.end(); ++i)
	cout << ' ' << *i;
    os << endl;
}

 bool stats_only = false; 
 bool silent = false; 

 unsigned long i_counter = 0;
 unsigned long f_counter = 0;
 unsigned long c_counter = 0;
 unsigned long m_counter = 0;
 unsigned long a_counter = 0;
 unsigned long x_counter = 0;
 unsigned long xx_counter = 0;

 Timer T;

 FwdTable FT;

 FTAllocator Mem(65536);

 static siena::string_t new_sx_string(const string & x) {
     char * s = new (Mem) char[x.length()];
     char * t = s;
     for(string::const_iterator c = x.begin(); c != x.end(); ++c)
	 *t++ = *c;
     return siena::string_t(s, t);
 }

%}

%pure_parser

/*
 * Keywords & other lex tokens
 */
%token <str> ID STR
%token <int_v> INT
%token <bool_v> BOOL
%token <double_v> DOUBLE
%token AND OR LT GT EQ NE PF SF SS
%token IFCONFIG SELECT 
%token PRINTSTATS STATSONLY TIMERSTART TIMERSTOP 
%token OUTPUTON OUTPUTOFF PREPROCESSLIMIT CLEAR CLEARRECYCLE
/*
 * some semantic types for non-terminals
 */
%type <value>		Value
%type <constraint>	Constraint
%type <message>		Message
%type <predicate>	Predicate
%type <op>		Op

%start StmtList

%%
/*
 * Grammar rules
 */
StmtList: Stmt | StmtList Stmt
;

Stmt: IfConfig | Select 
 | PRINTSTATS	{ sff_parser_print_statistics(); }
 | CLEAR	{ FT.clear(); }
 | CLEARRECYCLE { FT.clear_recycle(); }
 | STATSONLY	{ stats_only = true; }
 | TIMERSTART	{ T.start(); }
 | TIMERSTOP	{ T.stop(); }
 | OUTPUTON	{ silent = false; }
 | OUTPUTOFF	{ silent = true; }
 | PREPROCESSLIMIT EQ INT { FT.set_preprocess_rounds($3); }
;

IfConfig: IFCONFIG INT Predicate 
   {
       if (! stats_only) {
	   FT.ifconfig($2, *$3);
	   Mem.recycle();
       }
       ++i_counter;
   }
;

Predicate: ID Constraint 
   {
       if (! stats_only) {
	   $$ = new (Mem) simple_predicate();
	   $$->add(new (Mem) simple_filter());
	   $$->back()->add(new_sx_string($1), $2);
       }
       ++f_counter;
       ++c_counter;
   }
 | Predicate AND ID Constraint
   {
       if (! stats_only) {
	   $$ = $1;
	   $$->back()->add(new_sx_string($3), $4);
       }
       ++c_counter;
   }
 | Predicate OR ID Constraint
   {
       if (! stats_only) {
	   $$ = $1;
	   $$->add(new (Mem) simple_filter());
	   $$->back()->add(new_sx_string($3), $4);
       }
       ++f_counter;
       ++c_counter;
   }
;

Constraint: Op INT 
   { 
       if (!stats_only) { $$ = new (Mem) simple_op_value($1, $2); } 
   }
 | Op DOUBLE 
   { 
       if (!stats_only) { $$ = new (Mem) simple_op_value($1, $2); } 
   }
 | Op BOOL
   { 
       if (!stats_only) { $$ = new (Mem) simple_op_value($1, $2); } 
   }
 | Op STR
   { 
       if (!stats_only) { $$ = new (Mem) simple_op_value($1, 
							 new_sx_string($2)); } 
   }
;

Op: LT { $$ = siena::lt_id; }
 |  GT { $$ = siena::gt_id; }
 |  EQ { $$ = siena::eq_id; }
 |  PF { $$ = siena::pf_id; }
 |  SF { $$ = siena::sf_id; }
 |  SS { $$ = siena::ss_id; }
;

Value: INT { if (!stats_only) { $$ = new (Mem) simple_value($1); } }
 |  DOUBLE { if (!stats_only) { $$ = new (Mem) simple_value($1); } }
 |  BOOL   { if (!stats_only) { $$ =  new (Mem) simple_value($1); } }
 |  STR    { if (!stats_only) { $$ =  new (Mem) simple_value(new_sx_string($1)); } }
;

Select: SELECT Message 
   {
       if (! stats_only) {
	   if (silent) {
	       ICounter cc;
	       FT.match(*$2, cc);
	       if (cc.get_count() > 0) ++x_counter;
	       xx_counter += cc.get_count();
	   } else {
	       IPrinter ccc;
	       FT.match(*$2, ccc);
	       if (ccc.get_count() > 0) ++x_counter;
	       xx_counter += ccc.get_count();
	       ccc.flush(cout);
	   }
	   Mem.recycle();
	   ++m_counter;
       }
   }
;

Message: ID EQ Value
   { 
       if (! stats_only) {
	   $$ = new (Mem) simple_message();
	   $$->add(new_sx_string($1), $3);
       }
       ++a_counter;
   }
  | Message ID EQ Value
   {
       if (! stats_only) {
	   $$ = $1;
	   $$->add(new_sx_string($2), $4);
       }
       ++a_counter;
   }
;

%%

int yyerror(char *s) 
{
    cerr << yysfffname << ':' << yysffnlin << ": " << s << endl;
    return -1;
};

int sff_parser_run()
{
    sff_parser_clear_statistics();
    return yyparse();
}

int sff_parser_stats_only()
{
    int res;
    stats_only = true;
    res = yyparse();
    sff_parser_print_statistics();
    return res;
}

void sff_parser_timer_start() 
{
    T.start();
}

void sff_parser_timer_stop() 
{
    T.stop();
}

void sff_parser_print_statistics() 
{
    cout << "I=" << i_counter 
	 << " F=" << f_counter 
	 << " C=" << c_counter 
	 << " M=" << m_counter 
	 << " A=" << a_counter 
	 << " MX=" << x_counter 
	 << " MXX=" << xx_counter 
	 << " T=" << T.elapsed_usr_msec()
	 << " S=" << FT.bytesize()
	 << endl;
}

void sff_parser_clear_statistics() 
{
    i_counter = 0;
    f_counter = 0;
    c_counter = 0;
    m_counter = 0;
    a_counter = 0;
    x_counter = 0;
    xx_counter = 0;
}
