/*
 *  This file is part of Siena, a wide-area event notification system.
 *  See http://www.cs.colorado.edu/serl/siena/
 *  
 *  Author: Antonio Carzaniga <carzanig@cs.colorado.edu>
 *  See the file AUTHORS for full details. 
 *  
 *  Copyright (C) 1998-2002 University of Colorado
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
 *  USA, or send email to serl@cs.colorado.edu.
 *  
 *  
 *  $Id: sff_scanner.ll,v 1.2 2002/11/26 18:49:38 carzanig Exp $
 */
%{
#include <string>

#include "yysfftypes.h"
/* "yysfftypes.h" must be #included before "sff_parser.h" */
#include "sff_parser.h"

#define YY_NO_UNPUT

unsigned yysffnlin = 1; /* Current line number in source file */
string yysfffname = "{standard input}"; /* Current source file name */

int sff_scanner_open(const char * fname) {
    yysfffname = fname;
    yyin = fopen(fname, "r");
    return yyin == NULL;
}

int sff_scanner_close() {
    return fclose(yyin);
}

#define RET(x) {yylval->nlin = yysffnlin; yylval->fname = & yysfffname;};return(x)

%}

space			[ \b\t\v\f\r]+
alpha			[a-zA-Z_.\-]
nonzero			[1-9]
digit			[0-9]
oct_digit		[0-7]
hex_digit		[0-9a-fA-F]
ident			{alpha}({digit}|{alpha})*
nesc_seq		[^\\\"\'0-9ntvbrfax\?]
sign_opt		[\-+]?
decimal			({sign_opt}{nonzero}{digit}*)|("0")
octal			{oct_digit}{1,3}
hex			"0"[Xx]{hex_digit}+
exponent		[eE]{sign_opt}{digit}+
fractional		({digit}+".")|({digit}*"."{digit}+)
fpoint			{sign_opt}{fractional}{exponent}?
u			[Uu]
l			[Ll]
f			[Ff]

/* 
 *  mutually exclusive states 
 */
%x comment strlit 

%% 

\n#.*\n			{ const char * nlin; const char * fname;
                          if ((nlin = strtok(yytext, "\n#\t")) != NULL &&
                              (fname = strtok(NULL, " \t\"")) != NULL) {
			      yysffnlin = atoi(nlin);
			      yysfffname = fname;
                          }
                        }

"/*"			BEGIN(comment);
<comment>[^\*\n]*	;
<comment>\*[\*]*[^/\n]	;
<comment>\n		yysffnlin++;
<comment>[\*]*"*/"	BEGIN(INITIAL);

"//"[^\n]*\n		yysffnlin++;		

{space}			; 
\n			yysffnlin++;

\"			{ yylval->str.erase(yylval->str.begin(), 
					    yylval->str.end()); 
                          BEGIN(strlit); }
<strlit>[^\"\\\n]+	{ yylval->str.append(yytext); }
<strlit>\n		{ yylval->str.push_back('\n'); yysffnlin++; }
<strlit>\\n		yylval->str.push_back('\n');
<strlit>\\t		yylval->str.push_back('\t');
<strlit>\\r		yylval->str.push_back('\r');
<strlit>\\b		yylval->str.push_back('\b');
<strlit>\\f		yylval->str.push_back('\f');
<strlit>\\a		yylval->str.push_back('\a');
<strlit>\\v		yylval->str.push_back('\v');
<strlit>\\\"		yylval->str.push_back('\"');
<strlit>\\\'		yylval->str.push_back('\'');
<strlit>\\\?		yylval->str.push_back('\?');
<strlit>\\{nesc_seq}	; /* ignore non-escape character after "\" */
<strlit>\\{octal}	{ char *c = yytext; ++c; /* skip '\\' */
                          yylval->str.push_back((char)(strtoul(c, NULL, 8) % 256)); }
<strlit>\\{hex}		{ char *c = yytext; c += 3; /* skip "\\0x" */
                          yylval->str.push_back((char)(strtoul(c, NULL, 16) % 256)); }
<strlit>\"		{ BEGIN(INITIAL); RET(STR); }

","			RET(AND);
"|"			RET(OR);
"/\\"			RET(AND);
"\\/"			RET(OR);
"&&"			RET(AND);
"||"			RET(OR);
":"			RET(':');
"="			RET(EQ);
">"			RET(GT);
"<"			RET(LT);
"!="			RET(NE);
"=*"			RET(PF);
"*="			RET(SF);
"**"			RET(SS);

"ifconfig"		RET(IFCONFIG);
"select"		RET(SELECT);

"%print_statistics"	RET(PRINTSTATS);
"%statistics_only"	RET(STATSONLY);
"%timer_start"		RET(TIMERSTART);
"%timer_stop"		RET(TIMERSTOP);
"%output_on"		RET(OUTPUTON);
"%output_off"		RET(OUTPUTOFF);
"%pre_process_limit"	RET(PREPROCESSLIMIT);
"%clear"		RET(CLEAR);
"%clear_recycle"	RET(CLEARRECYCLE);

"true"			{ yylval->bool_v = true; RET(BOOL);}
"false"			{ yylval->bool_v = false; RET(BOOL);}

{ident}			{ yylval->str = yytext; RET(ID);}

{decimal}		{ yylval->int_v = atoi(yytext); RET(INT); }

{hex}			{ yylval->int_v = strtoul(yytext, NULL, 16); RET(INT); }

{fpoint}		{ yylval->double_v = strtod(yytext, NULL); RET(DOUBLE); }


%%

int yywrap() { return 1; };
